@@ -10,7 +10,12 @@
  * ----------------------------------------------------------------------- */
 
 #include "base.h"
-#include "linked_list.h"
+#include <linux/list.h>
+
+#ifdef __KERNEL__
+  #define malloc(x) kmalloc(x, GFP_KERNEL)
+  #define free kfree
+#endif
 
 #define LAYERS 8
 #define HASH_BITS 8
@@ -25,6 +30,12 @@ int translucent_fs    = 0;
 int translucent_flags = 0;
 int translucent_cnt   = 0;
 
+struct hash_data_t
+{
+	struct list_head list;
+	struct dirent64 data;
+};
+
 static inline int match_uids(void)
 {
    struct task_struct *p = current;
@@ -105,19 +116,28 @@ int translucent_create_whiteout(char *file) {
 	return 0;
 }
 
+/** create a copy of a regular file in overlay
+	@param nd specifies underlay - source to copy
+	@param nnew specifies new entry - destination
+	@return 0 on success, -ECODE otherwise
+*/
 int translucent_copy(struct nameidata *nd, struct nameidata *nnew, int lookup_flags) {
-	char *p, *buf;
+	char *p, *buf, *pathbuf;
 	ssize_t (*sys_write)(int fd, const void *buf, size_t count)=sys_call_table[__NR_write];	
 	ssize_t (*sys_read)(int fd, void *buf, size_t count)=sys_call_table[__NR_read];
 	int (*sys_close)(int fd)=sys_call_table[__NR_close];
-	int result,inphandle,outphandle,mode=nd->dentry->d_inode->i_mode;
+	int result,inphandle,outphandle;
+	int i_Bufsize=4096;
+	umode_t mode=nd->dentry->d_inode->i_mode;
+	struct utimbuf timebuf={ actime:nd->dentry->d_inode->i_atime, modtime:nd->dentry->d_inode->i_mtime };
 
 // exclude device/pipe/socket/dir and proc entries from COW
 	if(is_special(nd)) return -ENODEV;
 
 	mode &= S_IRWXUGO;
-        buf=malloc(REDIR_BUFSIZE+1);
-	p = d_path(nd->dentry, nd->mnt, buf, REDIR_BUFSIZE);
+        buf=malloc(i_Bufsize);
+        pathbuf=malloc(REDIR_BUFSIZE+1);
+	p = d_path(nd->dentry, nd->mnt, pathbuf, REDIR_BUFSIZE);
 
 //	printk(KERN_DEBUG SYSLOGID ": copy-on-write %s %o\n",p,mode);
 
@@ -127,7 +147,7 @@ int translucent_copy(struct nameidata *nd, struct nameidata *nnew, int lookup_fl
 	if(result<0) goto out_free;
 	inphandle=result;
 
-	p=d_path(nnew->dentry, nnew->mnt, buf, REDIR_BUFSIZE);
+	p=d_path(nnew->dentry, nnew->mnt, pathbuf, REDIR_BUFSIZE);
 	BEGIN_KMEM
 		result=orig_sys_open(p,O_WRONLY|O_CREAT,mode);
 	END_KMEM
@@ -136,17 +156,48 @@ int translucent_copy(struct nameidata *nd, struct nameidata *nnew, int lookup_fl
 
 	if(!(lookup_flags&LOOKUP_TRUNCATE)) {
 		BEGIN_KMEM
-			while((result=sys_read(inphandle,buf,REDIR_BUFSIZE))>0&&sys_write(outphandle,buf,result)>0);
+			while((result=sys_read(inphandle,buf,i_Bufsize))>0&&sys_write(outphandle,buf,result)>0);
 		END_KMEM
 	} else result=0;
 	sys_close(outphandle);
+	BEGIN_KMEM
+		orig_sys_utime(p, &timebuf);
+	END_KMEM
 out_close:
 	sys_close(inphandle);
 out_free:
         free(buf);
+        free(pathbuf);
 	return result;
 }
 
+/** create a copy of a symlink in overlay
+	@param nd specifies underlay - source to copy
+	@param nnew specifies new entry - destination
+	@return 0 on success, -ECODE otherwise
+*/
+int translucent_copylink(struct nameidata *nd, struct nameidata *nnew, int lookup_flags) {
+	char *p, *buf, *pathbuf;
+	int result;
+	buf=malloc(REDIR_BUFSIZE+1);
+	pathbuf=malloc(REDIR_BUFSIZE+1);
+	p=d_path(nd->dentry, nd->mnt, pathbuf, REDIR_BUFSIZE);
+	BEGIN_KMEM
+		result=orig_sys_readlink(p, buf, REDIR_BUFSIZE);
+	END_KMEM
+	if(result>=0) {
+		buf[result]=0;
+		p=d_path(nnew->dentry, nnew->mnt, pathbuf, REDIR_BUFSIZE);
+		BEGIN_KMEM
+			result=orig_sys_symlink(buf, p);
+		END_KMEM
+	}
+	free(pathbuf);
+	free(buf);
+	return result;
+}
+
+
 int translucent_mkdir(struct nameidata *nd, struct nameidata *n, int mode, struct translucent *t) {
 	char *p, *buf;
 	int result, umask=current->fs->umask;
@@ -175,7 +226,6 @@ static inline int mymkdir2(struct nameidata *nd, struct nameidata *n, struct tra
 	return translucent_mkdir(nd,n,nd->dentry->d_inode->i_mode,t);
 }
 
-unsigned int i_Random;
 /** calculate hash from key data
   @param unsigned char *key zero terminated input data
   @return int hash value in range [0,HASH_TABLE_SIZE-1]
@@ -189,7 +239,6 @@ static inline int hash_func(unsigned char *key)
                 hash+=*key;
                 ++key;
         }
-        i_Random^=hash;
         hash=(hash^(hash>>HASH_BITS)^(hash>>(HASH_BITS*2)))%HASH_TABLE_SIZE;
 //        printf("%i\n", hash);
         return hash;
@@ -206,25 +255,27 @@ static inline int translucent_getdents(unsigned int fd, struct dirent64 *dirp, u
 
 int translucent_merge_init(int i_Layers, struct nameidata *n)
 {
+	static unsigned int i_Random;
         int (*sys_close)(int)=sys_call_table[__NR_close];
         off_t (*sys_lseek)(int fildes, off_t offset, int whence)=sys_call_table[__NR_lseek];
         ssize_t (*sys_write)(int fd, const void *buf, size_t count)=sys_call_table[__NR_write];
         int (*sys_fchmod)(int fildes, mode_t mode)=sys_call_table[__NR_fchmod];
         int out, i, i_Len, i_Hash, i_Bytes, b_Found, b_Whiteout;
-        int i_HashSize=HASH_TABLE_SIZE*sizeof(linked_list_t *);
-        linked_list_t **hashtable, **p_CurList, *p_Cur;
-        void *p_Data;
+        int i_HashSize=HASH_TABLE_SIZE*sizeof(struct hash_data_t);
+        struct hash_data_t *hashtable, *p_CurList, *p_Cur;
+	struct list_head *ListHead, *ListHead2;
 	struct dirent64 *cur, *cur2, *dirents;
         int outfd, curdirfd, /*i_Valid=0, i_LastValid=0,*/ i_Layer;
-//        char buf[REDIR_BUFSIZE], s_File[REDIR_BUFSIZE];
 	char *p=NULL, *ps, *p_VarSpace, *buf, *s_File;
         struct task_struct *task;
         p_VarSpace=(char *)malloc(2*REDIR_BUFSIZE+2*sizeof(struct dirent64)+i_HashSize);
         buf=p_VarSpace; i=REDIR_BUFSIZE;
         s_File=p_VarSpace+i; i+=REDIR_BUFSIZE;
         dirents=(struct dirent64*)(p_VarSpace+i); i+=2*sizeof(struct dirent64);
-        hashtable=(linked_list_t **)(p_VarSpace+i);
-        memset(hashtable, 0, i_HashSize);
+        hashtable=(struct hash_data_t *)(p_VarSpace+i);
+        for(i=0; i<HASH_TABLE_SIZE; ++i) {
+		INIT_LIST_HEAD(&(hashtable[i].list));
+	}
         dirents[1].d_off=1;
 //        printk(SYSLOGID ": mi0\n");
 /*        for(i_Layer=0; i_Layer<i_Layers; ++i_Layer) {
@@ -274,22 +325,25 @@ int translucent_merge_init(int i_Layers, struct nameidata *n)
                                 b_Whiteout=translucent_is_whiteout2(s_File);
                                 i_Hash=hash_func(cur->d_name);
                                 b_Found=0;
-                                p_CurList=&(hashtable[i_Hash]);
-                                while((p_Cur=*p_CurList)) {
-                                        cur2=linked_list_get_data(p_Cur);
+                                p_CurList=hashtable+i_Hash;
+				list_for_each_safe(ListHead, ListHead2, (&p_CurList->list)) {
+                                        p_Cur=list_entry(ListHead, struct hash_data_t, list);
+					cur2=&(p_Cur->data);
                                         if(strcmp(cur2->d_name,cur->d_name)==0) {
                                                 if(b_Whiteout) {
-                                                        linked_list_remove(p_CurList);
+                                                        list_del(ListHead);
+							free(p_Cur);
                                                 }
                                                 b_Found=1;
                                                 break;
                                         }
-                                        p_CurList=(linked_list_t**)&(p_Cur->next);
                                 }
                                 if(!b_Found && !b_Whiteout){
-                                    p_Data=malloc(i_Len);
-                                    memcpy(p_Data, cur, i_Len);
-                                    linked_list_insert(&(hashtable[i_Hash]), p_Data);
+//				    p_Cur=malloc(sizeof(struct list_head)+i_Len+4); // save some space
+				    p_Cur=malloc(sizeof(struct hash_data_t));
+				    INIT_LIST_HEAD(&(p_Cur->list));
+                                    memcpy(&(p_Cur->data), cur, i_Len);
+                                    list_add_tail(&(p_Cur->list), &(p_CurList->list));
                                   }
 			        cur=(struct dirent64 *)((char *)cur+i_Len);
 		        }
@@ -299,8 +353,7 @@ int translucent_merge_init(int i_Layers, struct nameidata *n)
         }
         i_Bytes=0;
         task=current;
-        i_Random=(int)p_VarSpace^(int)task->pidhash_next^(int)task->pidhash_pprev^(int)&i^(int)sys_close^(i_Random<<1);
-        snprintf(s_File, REDIR_BUFSIZE, "/tmp/getdents-%.5i-%.8X", task->pid, i_Random);
+        snprintf(s_File, REDIR_BUFSIZE, "/tmp/getdents-%.5i-%.8X", task->pid, i_Random++);
         ps=p;
         for(i=strlen(s_File); ps && *ps && (unsigned)i<REDIR_BUFSIZE-1; ++i,++ps) {
                 s_File[i]=(*ps=='/')?'-':*ps; // append pathname with / substituted
@@ -317,17 +370,18 @@ int translucent_merge_init(int i_Layers, struct nameidata *n)
                 sys_write(outfd, &dirents[1], i_Len);
         END_KMEM
         for(i=0; i<HASH_TABLE_SIZE; ++i) {
-            linked_list_foreach(p_Cur,hashtable[i]) {
-                cur=(struct dirent64*)linked_list_get_data(p_Cur);
+            list_for_each_safe(ListHead,ListHead2,&(hashtable[i].list)) {
+	        p_Cur=list_entry(ListHead, struct hash_data_t, list);
+                cur=&(p_Cur->data);
                 //i_Len=cur->d_reclen; // space saving and a more human-readable file (strings /tmp/getdents...)
                 i_Bytes+=i_Len;
                 cur->d_off=i_Bytes;     // offset of next entry - needed to lseek in our `directory'
                 BEGIN_KMEM
                         sys_write(outfd, cur, i_Len);
                 END_KMEM
-                free(cur);
+		list_del(ListHead);
+                free(p_Cur);
             }
-            linked_list_finish(&hashtable[i]);
         }
         sys_lseek(outfd, 0, 0/*SEEK_SET*/);
         sys_fchmod(outfd, 01444);
@@ -472,6 +526,7 @@ int redirect_path_walk(char *name, char **endp,
 		                if(is_special(&n[j])) {
                                        if(S_ISDIR(mode) && (lflags&LOOKUP_CREATE)) mymkdir2(&n[j], &n[t->layers-1], t);
 			               else if((lflags&LOOKUP_NOSPECIAL)) error = -1; // no mknod on echo > /dev/null , cow on fifo and socket does not make sense, copying symlinks is unnecessary
+				       else if((lflags&LOOKUP_CREATE) && S_ISLNK(mode) && !(translucent_flags&no_copyonwrite)) error=translucent_copylink(&n[j],&n[t->layers-1],lflags);
 			               else if((lflags&LOOKUP_CREATE) && (S_ISBLK(mode) || S_ISCHR(mode)) && !(translucent_flags&no_copyonwrite)) {
 				               // this is inlined quasi "mymknod"
 				               char *buf=malloc(REDIR_BUFSIZE+1),*p=namei_to_path(&n[t->layers-1],buf);
